from pathlib import Path

def JsonFile(path: str | Path) -> Path:
    normalized = Path(path)

    # Ensure database files always use a .json suffix.
    if normalized.suffix.lower() != ".json":
        normalized = normalized.with_suffix(".json")

    return normalized


import asyncio
import time
from types import TracebackType
from typing import Type, TypeVar, Generic
from .file import JsonFile
from pathlib import Path
from pydantic import BaseModel, ValidationError

T = TypeVar("T", bound=BaseModel)

ASYNCIO_LOCKS_AVAILABLE: dict[str, asyncio.Lock] = {}

# Known Bugs/Features to Implement:
# - Atomic writes to prevent data loss. See: https://stackoverflow.com/a/2333872/11761617
# - Sync/async race, handle the case where a user uses both sync and async, which could lead to corruption.
# - Cross process protection, if multiple processes are accessing the same file, this could lead to corruption.
#    To fix, we would need to do OS level file locking.

# Dev Notes:
# - asyncio.to_thread is used to run the synchronous method in a separate thread,
#    this allows the event loop to remain working while the potentially blocking file I/O operation is performed.

class JsonDatabaseInit(Generic[T]):
    def __init__(
        self,
        path: str | Path,
        database_model: Type[T],
        backup_on_validation_error: bool,
    ):

        try:
            database_model()
        except ValidationError as error:
            raise ValueError(
                f"JsonDatabaseInit requires all fields to have defaults, but "
                f"{database_model.__name__} has required fields:\n{error}"
            )

        self.path = JsonFile(path)

        self.backup_on_validation_error = backup_on_validation_error
        self.database_model = database_model
        self.database: T | None = None
        
        self.lock_key = str(self.path.absolute())

    def get_lock(self) -> asyncio.Lock:
        lock = ASYNCIO_LOCKS_AVAILABLE.get(self.lock_key)

        if lock is None:
            lock = ASYNCIO_LOCKS_AVAILABLE[self.lock_key] = asyncio.Lock()

        return lock

    async def read_async(self) -> T:
        return await asyncio.to_thread(self.read)

    async def write_async(self, data: T) -> None:
        await asyncio.to_thread(self.write, data)

    async def __aenter__(self) -> T:
        lock = self.get_lock()
        await lock.acquire()

        self.database = await self.read_async()
        return self.database

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        try:
            if self.database is not None:
                await self.write_async(self.database)
        finally:
            lock = self.get_lock()
            lock.release()

    def write(self, data: T) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)

        self.path.write_text(
            data.model_dump_json(indent=4)
        )

    def backup(self) -> None:
        current_epoch_time = int(time.time())

        backup_path = self.path.with_name(f"{self.path.stem}_backup_{current_epoch_time}.json")
        
        backup_path.write_text(
            self.path.read_text()
        )

    def read(self) -> T:
        if not self.path.exists():
            self.write(self.database_model())
            return self.database_model()
        
        data = self.path.read_text()

        if data:
            try:
                return self.database_model.model_validate_json(data)
            except ValidationError as error:
                error_message = (
                    f"Failed to read database file {self.path} due to validation error:\n{error}\n"
                    "If this is for production use, consider implementing a migration strategy.\n"
                    "Consider: https://docs.pydantic.dev/2.9/concepts/validators/#model-validators"
                )

                if self.backup_on_validation_error:
                    self.backup()
                    error_message += (
                        "\nA backup of the invalid data has been created with a timestamp in the filename. "
                    )

                raise ValueError(error_message)
        else:
            self.write(self.database_model())
            return self.database_model()

    def __enter__(self) -> T:
        self.database = self.read()
        return self.database

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        if self.database is not None:
            self.write(self.database)

class JsonDatabaseOptional(Generic[T]):
    def __init__(
        self,
        path: str | Path,
        database_model: Type[T],
        backup_on_validation_error: bool,
    ):
        self.path = JsonFile(path)

        self.database_model = database_model
        self.backup_on_validation_error = backup_on_validation_error
        self.database: T | None = None
        self.lock_key = str(self.path.absolute())

    def get_lock(self) -> asyncio.Lock:
        lock = ASYNCIO_LOCKS_AVAILABLE.get(self.lock_key)

        if lock is None:
            lock = ASYNCIO_LOCKS_AVAILABLE[self.lock_key] = asyncio.Lock()

        return lock

    async def read_async(self) -> T | None:
        return await asyncio.to_thread(self.read)

    async def write_async(self, data: T) -> None:
        await asyncio.to_thread(self.write, data)

    async def __aenter__(self) -> T | None:
        lock = self.get_lock()
        await lock.acquire()
        self.database = await self.read_async()
        return self.database

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        try:
            if self.database is not None:
                await self.write_async(self.database)
        finally:
            lock = self.get_lock()
            lock.release()

    def write(self, data: T) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)

        self.path.write_text(
            data.model_dump_json(indent=4)
        )

    def backup(self) -> None:
        current_epoch_time = int(time.time())

        backup_path = self.path.with_name(f"{self.path.stem}_backup_{current_epoch_time}.json")
        
        backup_path.write_text(
            self.path.read_text()
        )

    def read(self) -> T | None:
        if not self.path.exists():
            return None

        data = self.path.read_text()

        if data:

            try:
                return self.database_model.model_validate_json(data)
            except ValidationError as error:
                error_message = (
                    f"Failed to read database file {self.path} due to validation error:\n{error}\n"
                    "If this is for production use, consider implementing a migration strategy.\n"
                    "Consider: https://docs.pydantic.dev/2.9/concepts/validators/#model-validators"
                )

                if self.backup_on_validation_error:
                    self.backup()
                    error_message += (
                        "\nA backup of the invalid data has been created with a timestamp in the filename. "
                    )

                raise ValueError(error_message)

        return None

    def __enter__(self) -> T | None:
        self.database = self.read()
        return self.database

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        if self.database is not None:
            self.write(self.database)


import subprocess
import sys
import config

# Dev Notes:
# - Referencing: https://semver.org/
# - `map` takes a function and some arguments, and applies the function to each of the arguments. 
#    It returns an iterator (a lazy generator) that you can convert to a list or iterate over directly.

def get_name(file_path: str) -> str:
    with open(file_path, "r") as f:
        for line in f:
            if line.startswith("name = "):
                name = line.split("=")[1].strip().strip('"')
                return name
    
    raise ValueError("Name not found in file")

def get_version(file_path: str) -> str:
    with open(file_path, "r") as f:
        for line in f:
            if line.startswith("version = "):
                version = line.split("=")[1].strip().strip('"')
                return version
    
    raise ValueError("Version not found in file")

def update_version(version: str, update_type: str) -> str:
    major, minor, patch = map(int, version.split("."))
    
    if update_type == "patch":
        patch += 1
    elif update_type == "minor":
        minor += 1
        patch = 0
    elif update_type == "major":
        major += 1
        minor = 0
        patch = 0

    return f'{major}.{minor}.{patch}'

def update_version_in_file(file_path: str, version_number: str) -> None:
    with open(file_path, "r") as f:
        lines = f.readlines()

    for i, line in enumerate(lines):
        if line.startswith("version = "):
            lines[i] = f'version = "{version_number}"\n'
            break

    with open(file_path, "w") as f:
        f.writelines(lines)

args =  sys.argv[1:]
version = get_version("pyproject.toml")
name = get_name("pyproject.toml")

if len(args) == 0:
    print(
        "No arguments provided\n",
        "Usage: python upload.py [patch|minor|major]\n",
        "Defaulting to patch update (incrementing the last number in the version)\n\n",
        'Reminder: Patch updates are for bug fixes and small improvements, "' \
        'minor updates are for new features that are backwards compatible, and major updates are for breaking changes.',
    )
    sys.exit(0)

if len(args) > 1:
    print(
        "Too many arguments, only one argument is allowed"
    )
    sys.exit(1)

update_type = args[0]

if update_type not in ["patch", "minor", "major"]:
    print(
        f"Invalid update type: {update_type}\n",
        "Usage: python upload.py [patch|minor|major]\n",
        'Reminder: Patch updates are for bug fixes and small improvements, "' \
        'minor updates are for new features that are backwards compatible, and major updates are for breaking changes.',
    )
    sys.exit(1)

new_version = update_version(version, update_type)
update_version_in_file("pyproject.toml", new_version)

subprocess.run(
    [sys.executable, "-m", "pip", "install", "hatchling", "twine"],
    capture_output=True,
)

result = subprocess.run(
    [sys.executable, "-m", "hatch", "build"],
    capture_output=True,
)

if result.returncode != 0:
    subprocess.run(
        [sys.executable, "-m", "pip", "install", "hatch"],
        capture_output=True,
    )

result = subprocess.run(
    [sys.executable, "-m", "hatch", "build"],
    capture_output=True,
)

if result.returncode != 0:
    print("Build failed:")
    print(result.stdout.decode())
    print(result.stderr.decode())

    # revert version change
    update_version_in_file("pyproject.toml", version)


result = subprocess.run(
    [
        sys.executable, "-m", "twine", "upload", "dist/*",
        "--username", "__token__",
        "--password", config.PYPI_API_TOKEN,
    ],
    capture_output=True,
)

if result.stderr:
    print("Upload failed:")
    print(result.stderr)
else:
    print("Upload successful!")
    print(f"You can find latest version here: https://pypi.org/project/{name}/{new_version}/")

# Now delete dist folder
result = subprocess.run(
    ["rm", "-rf", "dist"],
    capture_output=True,
)

if result.returncode != 0:
    print("Failed to delete dist folder:")
    print(result.stderr.decode())


from typing import Literal, Type, TypeVar, overload
from pathlib import Path
from .database import JsonDatabaseInit
from .database import JsonDatabaseOptional

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)

@overload
def JsonDatabase(
    path: str | Path, 
    models: Type[T], 
    auto_init: Literal[True], 
    backup_on_validation_error: bool = True,
) -> JsonDatabaseInit[T]: ...

@overload
def JsonDatabase(
    path: str | Path, 
    models: Type[T], 
    auto_init: Literal[False], 
    backup_on_validation_error: bool = True,
) -> JsonDatabaseOptional[T]: ...

@overload
def JsonDatabase(
    path: str | Path, 
    models: Type[T],
    *, # This forces the caller to use keyword arguments for the following parameters.
    backup_on_validation_error: bool = True,
) -> JsonDatabaseOptional[T]: ...

@overload
def JsonDatabase(
    path: str | Path, 
    models: Type[T],
) -> JsonDatabaseOptional[T]: ...

def JsonDatabase(
    path: str | Path,
    models: Type[T],
    auto_init: bool = False,
    backup_on_validation_error: bool = True,
) -> JsonDatabaseInit[T] | JsonDatabaseOptional[T]:
    if auto_init:
        return JsonDatabaseInit(
            path=path,
            database_model=models,
            backup_on_validation_error=backup_on_validation_error,
        )
    else:
        return JsonDatabaseOptional(
            path=path,
            database_model=models,
            backup_on_validation_error=backup_on_validation_error,
        )


from .file import JsonFile
from .factory import JsonDatabase
from .database import JsonDatabaseInit
from .database import JsonDatabaseOptional

# Controls what can be seen when importing with "from jays_tools.json_database import *"
# it'll always stem from the folder name, so since this __init__.py is in the "json_database" we
# have to specify from the starting point of the import, which is "jays_tools.json_database"
# & since there is no subfolder, we have to directly input from the file,
# which is why we can't do from jay_tool.json_database import JsonFile, JsonDatabase, JsonDatabaseInit, JsonDatabaseOptional
# instead we have to do from jays_tools.json_database.file import JsonFile, from jays_tools.json_database.factory import JsonDatabase, etc.

__all__ = [ 
    "JsonFile", 
    "JsonDatabase", 
    "JsonDatabaseInit", 
    "JsonDatabaseOptional"
]